#include "ethertap.h"
#include "network.h"
#include "tapnetbridge.h"
#include "sigint.h"
#include "main.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

_connectionType connectionType = ct_server;
bool verbose = false;
int tcpPort = -1;
char *host = 0;
char *password1 = 0, *password2 = 0;
int encryption = 0;
char *device = 0;

void parseParameters(int argc, char *argv[])
 {
  int c;
  bool quit = false;
  while ((c = getopt(argc, argv, "hicsm:p:l:r:e:d:v")) != -1)
   {
    switch (c)
     {
      case 'h' : 
       { 
        printf("Usage: %s -i -l <pwd1> -r <pwd2> -d <device> [-e <encryption>]\n", argv[0]);
        printf("       %s -c -l <pwd1> -r <pwd2> -d <device> -m <host> -p <port> [-e <encryption>] [-v]\n", argv[0]);
        printf("       %s -s -l <pwd1> -r <pwd2> -d <device> -p <port> [-e <encryption>] [-v]\n", argv[0]);
        printf("        -h : this help.\n");
        printf("        -i : INETD-mode.\n");
        printf("        -c : client-mode.\n");
        printf("        -s : server-mode.\n");
        printf("        -m : hostname of machine to connect to.\n");
        printf("        -p : TCP-port to use.\n");
        printf("        -l : local password.\n");
        printf("        -r : remote password.\n");
        printf("        -e : encryption (0: BLOWFISH; 1: DES; 2: 3DES; 3: 3WAY; 4: GOST; 6: SAFER64;\n");
        printf("                         7: SAFER128; 8: CAST128; 9: TEAN; 10: TWOFISH). \n");
        printf("        -d : ethertap device (e.g. /dev/tap0).\n");
        printf("        -v : verbose.\n");
        exit(0);
       }    
      case 'i' : connectionType = ct_inetd; break;
      case 'c' : connectionType = ct_client; break;
      case 's' : connectionType = ct_server; break;
      case 'm' : host = strcpy(new char[strlen(optarg)+1], optarg); break;
      case 'p' : tcpPort = atoi(optarg); break;
      case 'l' : password1 = strcpy(new char[strlen(optarg)+1], optarg); break;
      case 'r' : password2 = strcpy(new char[strlen(optarg)+1], optarg); break;
      case 'e' : encryption = atoi(optarg); break;
      case 'd' : device = strcpy(new char[strlen(optarg)+1], optarg); break;
      case 'v' : verbose = true; break;
      case '?' :  
       { 
        fprintf(stderr, "Unknown option. (-%c)\n", optopt);
        exit(1);
       }
     }
   }
  
  if ((connectionType == ct_client) && !host)
   {
    fprintf(stderr, "When acting as client, you must specify a host to connect to.\n");
    quit = true;
   }
  if ((connectionType != ct_client) && host)
   {
    fprintf(stderr, "You cannot specify a host, when acting as server.\n");
    quit = true;
   }
  if ((tcpPort != -1) && (connectionType == ct_inetd))
   {
    fprintf(stderr, "You cannot specify a TCP-port when using INETD-mode.\n");
    quit = true;
   } 
  if ((tcpPort == -1) && (connectionType != ct_inetd))
   {
    fprintf(stderr, "You have to specify a TCP-port when not in INETD-mode.\n");
    quit = true;
   }
  if (!password1 || !password2)
   {
    fprintf(stderr, "You must specify local and remote passwords.\n");
    quit = true;
   }
  if (!device)
   {
    fprintf(stderr, "You must specify an ethertap-device.\n");
    quit = true;
   } 
  if ((connectionType == ct_inetd) && verbose)
   {
    fprintf(stderr, "Verbose and INETD-support cannot be used together. (Would corrupt protocol.)\n");
    quit = true;
   }
   
  if (quit) exit(1); 
 }

int main(int argc, char *argv[])
 {
  ethertap *tap;
  network *net;
	tapNetBridge *bridge;

  parseParameters(argc, argv);
  
  srand(time(0));
  
  if (verbose)
   printf("TapTunnel %s: Copyright 1999/2000 by Lennart Poettering <lennart@poettering.de>.\nThis software is published under GPL.\n\n", VERSION); 
   
	while( 1 )
	{
		sigIntInitialize();
  try
   {
    tap = new ethertap(device); 
   }
  catch (ethertap::couldNotOpen)
   {
    fprintf(stderr, "Could not open ethertap-device. Did you compile your kernel with ethertap-support? Did you create the device-nodes? Did you load the ethertap-module?\n");
    return 1;
   }
  
  if (verbose) printf("TAP: Tap opened ('%s')\n", device);
  
  try
   {
    switch (connectionType)
     {
      case ct_inetd : net = new network(); break;
      case ct_server : net = new network(tcpPort); break;
      case ct_client : net = new network(host, tcpPort); break;
     }
   }
  catch (network::userBreak)
   {
    fprintf(stderr, "User break, canceled.\n");
    return 1;
   }      
  catch (network::couldNotOpen)
   {
    fprintf(stderr, "Could not create network connection.\n");
    return 1;
   }

  net->setLocalSecure(password1);
  net->setRemoteSecure(password2);
  
  try
   {
    net->setEncryption(encryption);
   }
  catch (network::unknownEncryption)
   {
    fprintf(stderr, "Unknown encryption.\n");
    return 1;
   }       
  
  try
   {
    net->handshake();
   }             
  catch (network::authFailure)
   {
    fprintf(stderr, "Authorization failed. Maybe your passwords are wrong.\n");
			goto cleanup;
   } 
  catch (network::handshakeFailure)
   {
    fprintf(stderr, "Handshake failed. Maybe something corrupted the data.\n");
			goto cleanup;
   } 
  catch (network::encryptionInconsistent)
   {
    fprintf(stderr, "Encryption inconsistent. You have to select the same encryption for both client and server.\n");
			goto cleanup;
   } 
  catch (network::cryptError)
   {
    fprintf(stderr, "Encryption engine failed to initialize.\n");
    return 1;
   }

  bridge = new tapNetBridge(tap, net);
  
  try
   {
    bridge->moveData();
   }
  catch (network::couldNotRead)  { fprintf(stderr, "Connection was reset by peer.\n"); }    
  catch (network::couldNotWrite) { fprintf(stderr, "Connection was reset by peer.\n"); }    
   
  cleanup:
   if (bridge) delete bridge;
   if (net) delete net;
   if (tap) delete tap;
   
  sigIntDeinitialize();
  }
  return 0; 
 }